import tempfile


registered_temporary_directories = []


class RegisteredTemporaryDirectory(tempfile.TemporaryDirectory):
    """A temporary directory that's registered at instantiation so it can be referenced later."""

    def __init__(self, suffix=None, prefix=None, dir=None, ignore_cleanup_errors=False):
        super().__init__(suffix=suffix, prefix=prefix, dir=dir, ignore_cleanup_errors=ignore_cleanup_errors)
        registered_temporary_directories.append(self)



import concurrent.futures
import copy
import datetime
import json
import logging
import os
from collections.abc import Iterable

import coolname
import requests

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import CloudLocationNotSpecified, InvalidInputException
from octue.mixins import CloudPathable, Hashable, Identifiable, Labelable, Metadata, Serialisable, Taggable
from octue.resources.datafile import Datafile
from octue.resources.filter_containers import FilterSet
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.files import RegisteredTemporaryDirectory
from octue.utils.metadata import METADATA_FILENAME, UpdateLocalMetadata, load_local_metadata_file


logger = logging.getLogger(__name__)


SIGNED_METADATA_DIRECTORY = ".signed_metadata_files"


class Dataset(Labelable, Taggable, Serialisable, Identifiable, Hashable, Metadata, CloudPathable):
    """A representation of a dataset with metadata.

    The default usage is to provide the path to a local or cloud directory and create the dataset from the files it
    contains. Alternatively, the `files` parameter can be provided and only those files are included. Either way, the
    `path` parameter should be explicitly set to something meaningful.

    Metadata consists of `id`, `name`, `tags`, and `labels`, available as attributes on the instance. On instantiation,
    metadata for the dataset is obtained from its stored location (the corresponding cloud object metadata
    or a local `.octue` metadata file) if present. Metadata values can alternatively be passed as arguments at
    instantiation but will only be used if stored metadata cannot be found - i.e. stored metadata always takes
    precedence (use the `ignore_stored_metadata` parameter to override this behaviour). Stored metadata can be updated
    after instantiation using the `update_metadata` method.

    :param str|None path: the path to the dataset (defaults to the current working directory if none is given)
    :param iter(str|dict|octue.resources.datafile.Datafile)|None files: the files belonging to the dataset
    :param bool recursive: if `True`, include in the dataset all files in the subdirectories recursively contained within the dataset directory
    :param bool ignore_stored_metadata: if `True`, ignore any metadata stored for this dataset locally or in the cloud and use whatever is given at instantiation
    :param bool include_octue_metadata_files: if `True`, include `.octue` metadata files as datafiles in the dataset when instantiating it
    :param str|None id: an optional UUID to assign to the dataset (defaults to a random UUID if none is given)
    :param str|None name: an optional name to give to the dataset (defaults to the dataset directory name)
    :param dict|octue.resources.tag.TagDict|None tags: key-value pairs with string keys conforming to the Octue tag format (see `TagDict`)
    :param iter(str)|octue.resources.label.LabelSet|None labels: space-separated string of labels relevant to the dataset
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("files",)
    _METADATA_ATTRIBUTES = ("id", "name", "tags", "labels")

    # Paths to files are added to the serialisation in `Dataset.to_primitive`.
    _SERIALISE_FIELDS = (*_METADATA_ATTRIBUTES, "path")

    def __init__(
        self,
        path=None,
        files=None,
        recursive=True,
        ignore_stored_metadata=False,
        include_octue_metadata_files=False,
        id=None,
        name=None,
        tags=None,
        labels=None,
    ):
        super().__init__(name=name, id=id, tags=tags, labels=labels)
        self.path = path or os.getcwd()
        self.files = FilterSet()
        self._recursive = recursive
        self._ignore_stored_metadata = ignore_stored_metadata
        self._include_octue_metadata_files = include_octue_metadata_files
        self._cloud_metadata = {}
        self._instantiated_from_files_argument = False

        if files:
            if not isinstance(files, Iterable) or isinstance(files, (dict, str)):
                raise InvalidInputException(
                    "The `files` parameter of a `Dataset` must be an iterable of `Datafile` instances, strings, or "
                    f"dictionaries. Received {files!r} instead."
                )

            self.files = self._instantiate_datafiles(files)
            self._instantiated_from_files_argument = True
            return

        if storage.path.is_cloud_path(self.path):
            self._instantiate_from_cloud(path=self.path)
            return

        self._instantiate_from_local_directory(path=self.path)

        if len(self.files) == 0:
            logger.warning("%r is empty at instantiation time (path %r).", self, self.path)

    @property
    def name(self):
        """Get the name of the dataset

        :return str:
        """
        if self._name:
            return self._name

        if self.exists_in_cloud:
            return storage.path.split(self.path)[-1].split("?")[0]

        return os.path.split(os.path.abspath(os.path.split(self.path)[-1]))[-1]

    @name.setter
    def name(self, name):
        """Set the name of the dataset.

        :param str name:
        :return None:
        """
        self._name = name

    @property
    def exists_locally(self):
        """Return `True` if the dataset exists locally.

        :return bool:
        """
        return not self.exists_in_cloud

    @property
    def all_files_are_in_cloud(self):
        """Do all the files of the dataset exist in the cloud?

        :return bool:
        """
        return all(file.exists_in_cloud for file in self.files)

    @property
    def metadata_path(self):
        """Get the path to the dataset's metadata file.

        :return str:
        """
        if self.exists_in_cloud:
            if storage.path.is_url(self.path):
                return self.path

            return storage.path.join(self.path, METADATA_FILENAME)

        return os.path.join(self.path, METADATA_FILENAME)

    def __iter__(self):
        yield from self.files

    def __len__(self):
        return len(self.files)

    def __contains__(self, item):
        return item in self.files

    def __enter__(self):
        """Enter the dataset metadata updating context.

        :return Dataset:
        """
        return self

    def __exit__(self, *args):
        """Update the cloud or local metadata for the dataset.

        :return None:
        """
        self.update_metadata()

    def upload(self, cloud_path=None, update_cloud_metadata=True):
        """Upload a dataset to the given cloud path.

        :param str|None cloud_path: cloud path to store dataset at (e.g. `gs://bucket_name/path/to/dataset`)
        :param bool update_cloud_metadata: if `True`, update the metadata of the dataset in the cloud at upload time
        :return str: cloud path for dataset
        """
        datafiles_and_paths = []

        for datafile in self.files:
            if self.exists_in_cloud:
                path_type = "cloud_path"
            else:
                path_type = "local_path"

            datafile_path_relative_to_dataset = self._datafile_path_relative_to_self(datafile, path_type=path_type)

            datafiles_and_paths.append(
                {
                    "datafile": datafile,
                    "new_cloud_path": storage.path.join(
                        cloud_path or self.path,
                        *datafile_path_relative_to_dataset.split(os.path.sep),
                    ),
                }
            )

        self._set_cloud_location(cloud_path)

        def upload_datafile(datafile_and_path):
            """Upload a datafile to the given cloud path.

            :param dict datafile_and_path:
            :return str:
            """
            return datafile_and_path["datafile"].upload(cloud_path=datafile_and_path["new_cloud_path"])

        # Use multiple threads to significantly speed up file uploads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for path in executor.map(upload_datafile, datafiles_and_paths):
                logger.debug("Uploaded datafile to %r.", path)

        if update_cloud_metadata:
            # If the dataset's metadata has been changed locally, update it in the cloud.
            local_metadata = self.metadata(use_octue_namespace=False)

            if self._cloud_metadata != local_metadata:
                self.update_cloud_metadata()

        return cloud_path

    def update_metadata(self):
        """Using the dataset instance's in-memory metadata, update its cloud metadata (if the dataset is cloud-based)
        or its local metadata file (if the dataset is local).

        :return None:
        """
        if self.exists_in_cloud:
            self.update_cloud_metadata()
            return

        self.update_local_metadata()

    def update_cloud_metadata(self):
        """Create or update the cloud metadata file for the dataset.

        :return None:
        """
        GoogleCloudStorageClient().upload_from_string(
            string=json.dumps({"dataset": self.to_primitive(include_files=False)}, cls=OctueJSONEncoder),
            cloud_path=self.metadata_path,
        )

    def update_local_metadata(self):
        """Create or update the local octue metadata file with the dataset's metadata.

        :return None:
        """
        with UpdateLocalMetadata(self) as existing_metadata_records:
            existing_metadata_records["dataset"] = self.to_primitive(include_files=False)
            os.makedirs(self.path, exist_ok=True)

    def generate_signed_url(self, expiration=datetime.timedelta(days=7)):
        """Generate a signed URL for the dataset. This is done by uploading a uniquely named metadata file containing
        signed URLs to the datasets' files and returning a signed URL to that metadata file.

        :param datetime.datetime|datetime.timedelta expiration: the amount of time or date after which the URL should expire
        :return str: the signed URL for the dataset
        """
        if not self.exists_in_cloud:
            raise CloudLocationNotSpecified(
                f"{self!r} must exist in the cloud for a signed URL to be generated for it."
            )

        signed_metadata = self.to_primitive()
        signed_metadata["files"] = [datafile.generate_signed_url(expiration=expiration) for datafile in self.files]

        path_to_signed_metadata_file = storage.path.join(self.path, SIGNED_METADATA_DIRECTORY, coolname.generate_slug())

        storage_client = GoogleCloudStorageClient()

        storage_client.upload_from_string(
            string=json.dumps(signed_metadata, cls=OctueJSONEncoder),
            cloud_path=path_to_signed_metadata_file,
        )

        return storage_client.generate_signed_url(cloud_path=path_to_signed_metadata_file, expiration=expiration)

    def add(self, datafile, path_in_dataset=None):
        """Add a datafile to the dataset. If the datafile's location is outside the dataset, it is copied to the dataset
        root or to the `path_in_dataset` if provided.

        :param octue.resources.datafile.Datafile datafile: the datafile to add to the dataset
        :param str|None path_in_dataset: if provided, set the datafile's local path to this path within the dataset
        :raise octue.exceptions.InvalidInputException: if the datafile is not a `Datafile` instance
        :return None:
        """
        if not isinstance(datafile, Datafile):
            raise InvalidInputException(f"{datafile!r} must be of type `Datafile` to add it to the dataset.")

        if self.exists_in_cloud:
            new_cloud_path = storage.path.join(self.path, path_in_dataset or datafile.name)

            # Add a cloud datafile to a cloud dataset.
            if datafile.exists_in_cloud:

                if datafile.cloud_path != new_cloud_path and not datafile.cloud_path.startswith(self.path):
                    datafile.upload(new_cloud_path)

                self.files.add(datafile)
                return

            # Add a local datafile to a cloud dataset.
            datafile.upload(new_cloud_path)
            self.files.add(datafile)
            return

        new_local_path = os.path.join(self.path, path_in_dataset or datafile.name)

        # Add a cloud datafile to a local dataset.
        if datafile.exists_in_cloud:
            datafile.download(local_path=new_local_path)
            self.files.add(datafile)
            return

        # Add a local datafile to a local dataset.
        if datafile.local_path != new_local_path and not datafile.local_path.startswith(self.path):
            datafile.local_path = new_local_path

        self.files.add(datafile)

    def get_file_by_label(self, label):
        """Get a single datafile from a dataset by filtering for files with the provided label.

        :param str label: the label to filter for
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if zero or more than one results satisfy the filters
        :return octue.resources.datafile.DataFile:
        """
        return self.files.one(labels__contains=label)

    def download(self, local_directory=None):
        """Download all files in the dataset.

        :param str|None local_directory: the path to a local directory to download the dataset into; if not provided, the files will be downloaded to a temporary directory
        :return str: the absolute path to the local directory
        """
        if not self.exists_in_cloud:
            raise CloudLocationNotSpecified(
                f"You can only download files from a cloud dataset. This dataset's path is {self.path!r}."
            )

        local_directory = os.path.abspath(local_directory or RegisteredTemporaryDirectory().name)
        datafiles_and_paths = []

        for file in self.files:

            if not file.exists_in_cloud:
                continue

            path_relative_to_dataset = self._datafile_path_relative_to_self(file, path_type="cloud_path")

            local_path = os.path.join(local_directory, *path_relative_to_dataset.split("/"))
            datafiles_and_paths.append({"datafile": file, "local_path": local_path})

        def download_datafile(datafile_and_path):
            """Download a datafile to the given path.

            :param dict datafile_and_path:
            :return str:
            """
            return datafile_and_path["datafile"].download(local_path=datafile_and_path["local_path"])

        # Use multiple threads to significantly speed up files downloads by reducing latency.
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for path in executor.map(download_datafile, datafiles_and_paths):
                logger.debug("Downloaded datafile to %r.", path)

        logger.info("Downloaded %r to %r.", self, local_directory)
        return local_directory

    def to_primitive(self, include_files=True):
        """Convert the dataset to a dictionary of primitives, converting its files into their paths for a lightweight
        serialisation.

        :param bool include_files: if `True`, include the `files` parameter in the dictionary
        :return dict:
        """
        serialised_dataset = super().to_primitive()

        if include_files:
            serialised_dataset["files"] = []

            for datafile in self.files:
                if datafile.exists_in_cloud:
                    datafile_path_type = "cloud_path"
                else:
                    datafile_path_type = "local_path"

                serialised_dataset["files"].append(getattr(datafile, datafile_path_type))

            serialised_dataset["files"] = sorted(serialised_dataset["files"])

        return serialised_dataset

    def _instantiate_from_cloud(self, path):
        """Instantiate the dataset from a cloud directory.

        :param str path: the cloud path to a directory in cloud storage
        :return None:
        """
        if not self._ignore_stored_metadata:
            self._use_cloud_metadata()

        # If files have been specified in the cloud metadata:
        if self.files:
            return

        # If no files have been specified in the cloud metadata:
        bucket_name = storage.path.split_bucket_name_from_cloud_path(path)[0]

        if self._include_octue_metadata_files:
            filter = lambda blob: SIGNED_METADATA_DIRECTORY not in blob.name
        else:
            filter = lambda blob: (
                SIGNED_METADATA_DIRECTORY not in blob.name and not blob.name.endswith(METADATA_FILENAME)
            )

        self.files = FilterSet(
            Datafile(
                path=storage.path.generate_gs_path(bucket_name, blob.name),
                ignore_stored_metadata=self._ignore_stored_metadata,
            )
            for blob in GoogleCloudStorageClient().scandir(path, recursive=self._recursive, filter=filter)
        )

    def _instantiate_from_local_directory(self, path):
        """Instantiate the dataset from a local directory.

        :param str path: the path to a local directory
        :return None:
        """
        self.files = FilterSet()

        for level, (directory_path, _, filenames) in enumerate(os.walk(path)):
            if not self._recursive and level > 0:
                break

            for filename in filenames:
                if filename == METADATA_FILENAME and not self._include_octue_metadata_files:
                    continue

                self.files.add(
                    Datafile(
                        path=os.path.join(directory_path, filename),
                        ignore_stored_metadata=self._ignore_stored_metadata,
                    )
                )

        if not self._ignore_stored_metadata:
            self._use_local_metadata()

    def _instantiate_datafiles(self, files):
        """Instantiate and add the given files to a `FilterSet`.

        :param iter(str|dict|octue.resources.datafile.Datafile) files:
        :return octue.resources.filter_containers.FilterSet:
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            return FilterSet(executor.map(self._instantiate_datafile, copy.deepcopy(files)))

    def _instantiate_datafile(self, file):
        """Instantiate a datafile from multiple input formats.

        :param str|dict|octue.resources.datafile.Datafile file:
        :return octue.resources.datafile.Datafile:
        """
        if isinstance(file, Datafile):
            return file

        if isinstance(file, str):
            return Datafile(path=file, ignore_stored_metadata=self._ignore_stored_metadata)

        return Datafile.deserialise(file)

    def _get_cloud_metadata(self):
        """Get the cloud metadata for the given dataset if a dataset metadata file has previously been uploaded.

        :return None:
        """
        if storage.path.is_url(self.path):
            try:
                self._cloud_metadata = requests.get(self.path).json()
            except requests.exceptions.ConnectionError:
                pass
            return

        storage_client = GoogleCloudStorageClient()

        if not storage_client.exists(cloud_path=self.metadata_path):
            return

        self._cloud_metadata = json.loads(storage_client.download_as_string(cloud_path=self.metadata_path)).get(
            "dataset", {}
        )

    def _use_cloud_metadata(self):
        """Update the dataset instance's metadata from the metadata file located in its cloud directory. If no metadata
        is stored for the dataset, do nothing.

        :return None:
        """
        self._get_cloud_metadata()

        if not self._cloud_metadata:
            return

        self._set_metadata(self._cloud_metadata)

    def _use_local_metadata(self):
        """Update the dataset instance's metadata from the local metadata records file. If no metadata is stored for the
        dataset, do nothing.

        :return None:
        """
        local_metadata = load_local_metadata_file(self)
        dataset_metadata = local_metadata.get("dataset", {})

        if not dataset_metadata:
            return

        self._set_metadata(dataset_metadata)

    def _set_metadata(self, metadata):
        """Set the dataset's metadata.

        :param dict metadata:
        :return None:
        """
        if "files" in metadata:
            # There's no need to provide the `ignore_stored_metadata` parameter here as this method is only used when
            # stored metadata is used.
            self.files = FilterSet(Datafile(path=path) for path in metadata["files"])

        for attribute in self._METADATA_ATTRIBUTES:
            if attribute not in metadata:
                continue

            if attribute == "id":
                self._set_id(metadata["id"])
                continue

            setattr(self, attribute, metadata[attribute])

    def _datafile_path_relative_to_self(self, datafile, path_type):
        """Get the path of the given datafile relative to the dataset.

        :param octue.resources.datafile.Datafile datafile: the datafile to get the relative path for
        :param str path_type: the datafile path type to use to calculate the relative path - one of "cloud_path" or "local_path"
        :return str: the relative path
        """
        if storage.path.is_url(self.path):
            dataset_path = self.path.split(SIGNED_METADATA_DIRECTORY)[0].strip("/")
        else:
            dataset_path = self.path

        datafile_path = getattr(datafile, path_type)

        if storage.path.is_url(datafile_path):
            datafile_path = datafile_path.split("?")[0]

        return storage.path.relpath(datafile_path, dataset_path)



import copy
import json
import logging
import os

from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


cached_local_metadata_files = {}
METADATA_FILENAME = ".octue"


class UpdateLocalMetadata:
    """A context manager that provides the contents of the given dataset's or datafile's local metadata file and updates
    it with any changes made within its context. The local metadata is retrieved either from the disk or from the cache
    as appropriate.

    :param octue.resources.datafile.Datafile|octue.resources.dataset.Dataset datafile_or_dataset: the datafile or dataset to update the local metadata for
    :return None:
    """

    def __init__(self, datafile_or_dataset):
        self.datafile_or_dataset = datafile_or_dataset
        self._local_metadata = None

    def __enter__(self):
        """Load the local metadata file and return its contents.

        :return any: the contents of the local metadata file (converted from the JSON in the local metadata file)
        """
        self._local_metadata = load_local_metadata_file(self.datafile_or_dataset)
        return self._local_metadata

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Write any changes to the contents of the file.

        :return None:
        """
        overwrite_local_metadata_file(self._local_metadata, self.datafile_or_dataset)


def load_local_metadata_file(datafile_or_dataset):
    """Load metadata from the local metadata records file for a datafile or dataset, returning an empty dictionary if
    the file does not exist or is incorrectly formatted. If the file has already been cached, its contents are retrieved
    from the cache.

    :param octue.resources.datafile.Datafile|octue.resources.dataset.Dataset datafile_or_dataset: the datafile or dataset to load the local metadata file for
    :return dict: the contents of the local metadata file
    """
    cached_metadata = _get_metadata_from_cache(datafile_or_dataset)

    if cached_metadata:
        return cached_metadata

    if not os.path.exists(datafile_or_dataset.metadata_path):
        local_metadata = {}

    else:
        with open(datafile_or_dataset.metadata_path) as f:
            try:
                local_metadata = json.load(f, cls=OctueJSONDecoder)
            except json.decoder.JSONDecodeError:
                logger.warning(
                    f"The metadata file at {datafile_or_dataset.metadata_path!r} is incorrectly formatted so no "
                    f"metadata can be read from it. Please fix or delete it."
                )
                local_metadata = {}

    _overwrite_cache_entry(datafile_or_dataset, local_metadata)
    return local_metadata


def overwrite_local_metadata_file(data, datafile_or_dataset):
    """Create or overwrite the local metadata file of a datafile or dataset with the given data. If the data to
    overwrite the file with is the same as the file's cache entry, no changes are made. If it's not, the cache entry is
    updated and the file is overwritten.

    :param dict data: the data to overwrite the local metadata file with
    :param octue.resources.datafile.Datafile|octue.resources.dataset.Dataset datafile_or_dataset: the datafile or dataset to overwrite the local metadata file for
    :return None:
    """
    cached_metadata = _get_metadata_from_cache(datafile_or_dataset)

    if data == cached_metadata:
        logger.debug("Avoiding overwriting local metadata file - its data is already in sync with the cache.")
        return

    _overwrite_cache_entry(datafile_or_dataset, data)

    with open(datafile_or_dataset.metadata_path, "w") as f:
        json.dump(data, f, cls=OctueJSONEncoder, indent=4)
        f.write("\n")


def _get_metadata_from_cache(datafile_or_dataset):
    """Get metadata for a datafile or dataset from the cache. If it's not cached, return `None`.

    :param octue.resources.datafile.Datafile|octue.resources.dataset.Dataset datafile_or_dataset: the datafile or dataset to get metadata from the cache for
    :return dict|None: the metadata or, if the file hasn't been cached, `None`
    """
    logger.debug("Using cached local metadata for %r.", datafile_or_dataset)
    return copy.deepcopy(cached_local_metadata_files.get(datafile_or_dataset.id))


def _overwrite_cache_entry(datafile_or_dataset, data):
    """Overwrite the metadata cache entry for a datafile or dataset.

    :param octue.resources.datafile.Datafile|octue.resources.dataset.Dataset datafile_or_dataset: the datafile or dataset to overwrite metadata in the cache for
    :param dict data: the data to overwrite the existing cache entry with.
    :return None:
    """
    cached_local_metadata_files[datafile_or_dataset.id] = copy.deepcopy(data)
    logger.debug("Updated local metadata cache for %r.", datafile_or_dataset)



import json

from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder


class Serialisable:
    """A mixin class to make instances serialisable as JSON.

    When calling ``Serialisable.serialise``, by default, only public attributes are included. If the class variable
    ``_SERIALISE_FIELDS`` is specified, then only these exact attributes are included (these can include non-public
    attributes); conversely, if ``_EXCLUDE_SERIALISE_FIELDS`` is specified and ``_SERIALISE_FIELDS`` is not, then all
    public attributes are included apart from the excluded ones.

    By default, JSON conversion is carried out using the ``OctueJSONEncoder``, which will sort keys as well as format
    and indent automatically. Additional keyword arguments will be passed to ``json.dumps`` to allow other formatting
    options.

    For example:

    .. code-block:: python

        class MyThing(Serialisable):

            _SERIALISE_FIELDS = ("a",)

            def __init__(self):
                self.a = 1
                self.b = 2
                self._private = 3
                self.__protected = 4

        MyThing().to_primitive()
        >>> {"a": 1}

    """

    _SERIALISE_FIELDS = None
    _EXCLUDE_SERIALISE_FIELDS = tuple()

    @classmethod
    def deserialise(cls, serialised_object, from_string=False):
        """Deserialise the given JSON-serialised object into an instance of the class.

        :param str|dict serialised_object: the string or dictionary of python primitives to deserialise into an instance
        :param bool from_string: if ``True``, deserialise from a JSON string; otherwise, deserialise from a dictionary
        :return any: an instance of the class
        """
        if from_string:
            serialised_object = json.loads(serialised_object, cls=OctueJSONDecoder)

        return cls(**serialised_object)

    @classmethod
    def from_file(cls, path, **kwargs):
        """Deserialise an instance from the given file.

        :param str path: the path to the JSON file containing the serialised instance
        :param kwargs: kwargs to pass in to the JSON deserialisation
        :return any: an instance of the class
        """
        with open(path) as f:
            return cls.deserialise(json.load(f, cls=OctueJSONDecoder, **kwargs))

    def serialise(self, **kwargs):
        """Serialise the instance to a JSON string of primitives. See the ``Serialisable`` constructor for more
        information.

        :param kwargs: kwargs to pass in to the JSON serialisation
        :return str: a JSON string containing the instance as a serialised python primitive
        """
        return json.dumps(self.to_primitive(), cls=OctueJSONEncoder, sort_keys=True, indent=4, **kwargs)

    def to_primitive(self):
        """Convert the instance into a JSON-compatible python dictionary of its attributes as primitives. See the
        ``Serialisable`` constructor for more information.

        :return dict:
        """
        names_of_attributes_to_serialise = self._SERIALISE_FIELDS or (
            field_name
            for field_name in dir(self)
            if (
                field_name not in self._EXCLUDE_SERIALISE_FIELDS
                and (field_name[:1] != "_")
                and (type(getattr(self, field_name, "")).__name__ != "method")
            )
        )

        self_as_primitive = {}

        for name in names_of_attributes_to_serialise:
            attribute = getattr(self, name, None)

            # Serialise sets as sorted list (JSON doesn't support sets).
            if isinstance(attribute, set):
                self_as_primitive[name] = sorted(attribute)
            else:
                self_as_primitive[name] = attribute

        return self_as_primitive

    def to_file(self, path, **kwargs):
        """Write the instance to a JSON file.

        :param str path: path of file to write to, including relative or absolute path and .json extension
        :param kwargs: kwargs to pass in to the JSON serialisation
        :return None:
        """
        with open(path, "w") as f:
            f.write(self.serialise(**kwargs))



import uuid


def gen_uuid():
    """Generate a unique identifier for an object.
    TODO - generate using an Octue api call, so we can register and find objects later using their UUID

    :return str:
    """
    return str(uuid.uuid4())



import logging
import os

from .log_handlers import apply_log_handler, should_use_octue_log_handler
from .runner import Runner


logger = logging.getLogger(__name__)


PYTHONUNBUFFERED = "PYTHONUNBUFFERED"
__all__ = ("Runner",)


REPOSITORY_ROOT = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


if should_use_octue_log_handler():
    apply_log_handler(
        logger_name=None,  # Apply to the root logger.
        include_line_number=bool(int(os.environ.get("INCLUDE_LINE_NUMBER_IN_LOGS", 0))),
        include_process_name=bool(int(os.environ.get("INCLUDE_PROCESS_NAME_IN_LOGS", 0))),
        include_thread_name=bool(int(os.environ.get("INCLUDE_THREAD_NAME_IN_LOGS", 0))),
    )

    if not os.environ.get(PYTHONUNBUFFERED):
        logger.warning(
            f"The {PYTHONUNBUFFERED!r} environment variable isn't set - logs may not appear in real time. Set "
            f"{PYTHONUNBUFFERED}=1 to fix this."
        )



class MultiPatcher:
    """A patch context manager that applies multiple patches at once. All the given patches are started on enter and
    stopped on exit.

    :param list(unittest.mock._patch) patches:
    :return None:
    """

    def __init__(self, patches=None):
        self.patches = patches or []

    def __enter__(self):
        """Start the patches and return the mocks they produce.

        :return list(unittest.mock.MagicMock):
        """
        return self.start()

    def __exit__(self, *args, **kwargs):
        """Stop the patches.

        :return None:
        """
        self.stop()

    def start(self):
        """Start the patches and return the mocks they produce.

        :return list(unittest.mock.MagicMock):
        """
        return [patch.start() for patch in self.patches]

    def stop(self):
        """Stop the patches.

        :return None:
        """
        for patch in self.patches:
            patch.stop()



import logging

import google.api_core.exceptions
from flask import Flask, request

from octue.cloud.deployment.google.answer_pub_sub_question import answer_question
from octue.cloud.pub_sub.bigquery import get_events
from octue.configuration import ServiceConfiguration


logger = logging.getLogger(__name__)
app = Flask(__name__)


QUESTION_ACKNOWLEDGMENT_RESPONSE = ("", 204)


@app.route("/", methods=["POST"])
def index():
    """Receive questions from Google Cloud Run in the form of Google Pub/Sub messages.

    :return (str, int):
    """
    envelope = request.get_json()

    if not envelope:
        return _log_bad_request_and_return_400_response("No Pub/Sub message received.")

    if not isinstance(envelope, dict) or "message" not in envelope:
        return _log_bad_request_and_return_400_response(f"Invalid Pub/Sub message format - received {envelope!r}.")

    question = envelope["message"]

    if "data" not in question or "attributes" not in question or "question_uuid" not in question["attributes"]:
        return _log_bad_request_and_return_400_response(f"Invalid Pub/Sub message format - received {envelope!r}.")

    should_drop_question = _check_if_should_drop_question(
        question_uuid=question["attributes"].get("question_uuid"),
        retry_count=question["attributes"].get("retry_count"),
    )

    if should_drop_question:
        return QUESTION_ACKNOWLEDGMENT_RESPONSE

    project_name = envelope["subscription"].split("/")[1]
    answer_question(question=question, project_name=project_name)
    return QUESTION_ACKNOWLEDGMENT_RESPONSE


def _log_bad_request_and_return_400_response(message):
    """Log an error return a bad request (400) response.

    :param str message:
    :return (str, int):
    """
    logger.error(message)
    return (f"Bad Request: {message}", 400)


def _check_if_should_drop_question(question_uuid, retry_count):
    """Check if the question has been delivered before and should be dropped. If it's a new question or if the question
    is an explicit retry (i.e. it's been received before but the retry count is unique), return `False`.

    To determine if a question is new, the event store is checked for a delivery acknowledgement for the question UUID.
    If no event store is specified or the specified event store can't be found, return `False`.

    :param str question_uuid: the UUID of the question to check
    :param str retry_count: the retry count of the question to check (an integer in string form)
    :return bool: whether the question should be dropped
    """
    service_configuration = ServiceConfiguration.from_file()

    if not service_configuration.event_store_table_id:
        logger.warning(
            "Cannot check if question has been redelivered as the 'event_store_table_id' key hasn't been set in the "
            "service configuration (`octue.yaml` file)."
        )
        return False

    try:
        previous_question_attempts = get_events(
            table_id=service_configuration.event_store_table_id,
            question_uuid=question_uuid,
            kind="delivery_acknowledgement",
        )

    except google.api_core.exceptions.NotFound:
        logger.warning(
            "Cannot check if question has been redelivered as no event store table was found with the ID %r; check "
            "that the 'event_store_table_id' key in the service configuration (`octue.yaml` file) is correct.",
            service_configuration.event_store_table_id,
        )
        return False

    # If there are no events for this question UUID, assume this is the first attempt for the question.
    if not previous_question_attempts:
        logger.info("Question %r (retry count %s) is a new question.", question_uuid, retry_count)
        return False

    # Acknowledge redelivered questions to stop further redundant redelivery and processing.
    for event in previous_question_attempts:
        if event["attributes"]["retry_count"] == retry_count:
            logger.warning(
                "Question %r (retry count %s) has already been received by the service. It will now be acknowledged "
                "and dropped to prevent further redundant redelivery.",
                question_uuid,
                retry_count,
            )
            return True

    return False



import logging

from octue.cloud.events.handler import AbstractEventHandler
from octue.cloud.events.validation import SERVICE_COMMUNICATION_SCHEMA


logger = logging.getLogger(__name__)


class EventReplayer(AbstractEventHandler):
    """A replayer for events retrieved asynchronously from storage.

    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param bool record_events: if `True`, record received events in the `received_events` attribute
    :param dict|None event_handlers: a mapping of event type names to callables that handle each type of event. The handlers must not mutate the events.
    :param dict|str schema: the JSON schema to validate events against
    :param bool include_service_metadata_in_logs: if `True`, include the SRUIDs and question UUIDs of the service revisions involved in the question to the start of the log message
    :param bool only_handle_result: if `True`, skip non-result events and only handle the "result" event if present
    :return None:
    """

    def __init__(
        self,
        handle_monitor_message=None,
        record_events=True,
        event_handlers=None,
        schema=SERVICE_COMMUNICATION_SCHEMA,
        include_service_metadata_in_logs=True,
        only_handle_result=False,
    ):
        event_handlers = event_handlers or {
            "question": self._handle_question,
            "delivery_acknowledgement": self._handle_delivery_acknowledgement,
            "heartbeat": self._handle_heartbeat,
            "monitor_message": self._handle_monitor_message,
            "log_record": self._handle_log_message,
            "exception": self._handle_exception,
            "result": self._handle_result,
        }

        super().__init__(
            handle_monitor_message=handle_monitor_message,
            record_events=record_events,
            event_handlers=event_handlers,
            schema=schema,
            include_service_metadata_in_logs=include_service_metadata_in_logs,
            only_handle_result=only_handle_result,
        )

    def handle_events(self, events):
        """Handle the given events and return a handled "result" event if one is present.

        :param iter(dict) events: the events to handle
        :return dict|None: the handled "result" event if present
        """
        super().handle_events()

        for event in events:
            event, attributes = self._extract_and_validate_event(event)

            # Skip the event if it fails validation.
            if not event:
                continue

            result = self._handle_event(event, attributes)

            if result:
                return result

    def _extract_event_and_attributes(self, container):
        """Extract an event and its attributes from the event container.

        :param dict container: the container of the event
        :return (any, dict): the event and its attributes
        """
        return container["event"], container["attributes"]

    def _handle_question(self, event, attributes):
        """Log that the question was sent.

        :param dict event:
        :param dict attributes: the event's attributes
        :return None:
        """
        logger.info(
            "%r asked a question %r to service %r.",
            attributes["sender"],
            attributes["question_uuid"],
            attributes["recipient"],
        )



from octue.cloud.events import OCTUE_SERVICES_PREFIX
from octue.cloud.service_id import convert_service_id_to_pub_sub_form

from .subscription import Subscription
from .topic import Topic


__all__ = ["Subscription", "Topic"]


def create_push_subscription(
    project_name,
    sruid,
    push_endpoint,
    subscription_filter=None,
    expiration_time=None,
    allow_existing=True,
):
    """Create a Google Pub/Sub push subscription for an Octue service for it to receive questions from parents. If a
    corresponding topic doesn't exist, it will be created first.

    :param str project_name: the name of the Google Cloud project in which the subscription will be created
    :param str sruid: the SRUID (service revision unique identifier)
    :param str push_endpoint: the HTTP/HTTPS endpoint of the service to push to. It should be fully formed and include the 'https://' prefix
    :param str|None subscription_filter: if specified, the filter to apply to the subscription; otherwise, no filter is applied
    :param float|None expiration_time: the number of seconds of inactivity after which the subscription should expire. If not provided, no expiration time is applied to the subscription
    :param bool allow_existing: if True, don't raise an error if the subscription already exists
    :return octue.cloud.pub_sub.subscription.Subscription:
    """
    if expiration_time:
        expiration_time = float(expiration_time)
    else:
        expiration_time = None

    subscription = Subscription(
        name=convert_service_id_to_pub_sub_form(sruid),
        topic=Topic(name=OCTUE_SERVICES_PREFIX, project_name=project_name),
        filter=subscription_filter,
        expiration_time=expiration_time,
        push_endpoint=push_endpoint,
    )

    subscription.create(allow_existing=allow_existing)
    return subscription



import concurrent.futures
import copy
import logging
import os
import uuid

from octue.cloud.pub_sub.service import Service
from octue.resources import service_backends


logger = logging.getLogger(__name__)

BACKEND_TO_SERVICE_MAPPING = {"GCPPubSubBackend": Service}


class Child:
    """A class representing an Octue child service that can be asked questions. This is a convenience wrapper for
    `Service` that makes asking questions more intuitive and allows easier selection of backends.

    :param str id: the ID of the child
    :param dict backend: must include the key "name" with a value of the name of the type of backend e.g. "GCPPubSubBackend" and key-value pairs for any other parameters the chosen backend expects
    :param str internal_sruid: the SRUID to give to the internal service used to ask questions to the child
    :param iter(dict)|None service_registries: the names and endpoints of the registries used to resolve the child's service revision when asking it questions; these should be in priority order (highest priority first)
    :return None:
    """

    def __init__(self, id, backend, internal_sruid="local/local:local", service_registries=None):
        self.id = id

        backend = copy.deepcopy(backend)
        backend_type_name = backend.pop("name")
        backend = service_backends.get_backend(backend_type_name)(**backend)

        self._service = BACKEND_TO_SERVICE_MAPPING[backend_type_name](
            service_id=internal_sruid,
            backend=backend,
            service_registries=service_registries,
        )

    def __repr__(self):
        """Represent the child as a string.

        :return str:
        """
        return f"<{type(self).__name__}({self.id!r})>"

    @property
    def received_events(self):
        """Get the events received from the child if it has been asked a question. If it hasn't, `None` is returned.
        If an empty list is returned, no messages have been received.

        :return list(dict)|None:
        """
        return self._service.received_events

    def ask(
        self,
        input_values=None,
        input_manifest=None,
        children=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        handle_monitor_message=None,
        record_events=True,
        save_diagnostics="SAVE_DIAGNOSTICS_ON_CRASH",  # This is repeated as a string here to avoid a circular import.
        question_uuid=None,
        parent_question_uuid=None,
        originator_question_uuid=None,
        originator=None,
        push_endpoint=None,
        asynchronous=False,
        retry_count=0,
        timeout=86400,
        maximum_heartbeat_interval=300,
    ):
        """Ask the child either:
        - A synchronous (ask-and-wait) question and wait for it to return an output. Questions are synchronous if
          the `push_endpoint` isn't provided and `asynchronous=False`.
        - An asynchronous (fire-and-forget) question and return immediately. To make a question asynchronous, provide
          the `push_endpoint` argument or set `asynchronous=True`.

        :param any|None input_values: any input values for the question, conforming with the schema in the child's twine
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question, conforming with the schema in the child's twine
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys.
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the child and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the child will have access to these local files
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_events: if `True`, record messages received from the child in the `received_events` property
        :param str save_diagnostics: must be one of {"SAVE_DIAGNOSTICS_OFF", "SAVE_DIAGNOSTICS_ON_CRASH", "SAVE_DIAGNOSTICS_ON"}; if turned on, allow the input values and manifest (and its datasets) to be saved by the child either all the time or just if it fails while processing them
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :param str|None parent_question_uuid: the UUID of the question that triggered this question
        :param str|None originator_question_uuid: the UUID of the question that triggered all ancestor questions of this question; if `None`, this question is assumed to be the originator question
        :param str|None originator: the SRUID of the service revision that triggered all ancestor questions of this question;  if `None`, this service revision is assumed to be the originator
        :param str|None push_endpoint: if answers to the question should be pushed to an endpoint, provide its URL here (the returned subscription will be a push subscription); if not, leave this as `None`
        :param bool asynchronous: if `True`, don't wait for an answer or create an answer subscription (the result and other events can be retrieved from the event store later)
        :param int retry_count: the retry count of the question (this is zero if it's the first attempt at the question)
        :param float timeout: time in seconds to wait for an answer before raising a timeout error
        :param float|int maximum_heartbeat_interval: the maximum amount of time (in seconds) allowed between child heartbeats before an error is raised
        :raise TimeoutError: if the timeout is exceeded while waiting for an answer
        :return dict|octue.cloud.pub_sub.subscription.Subscription|None, str: for a synchronous question, a dictionary containing the keys "output_values" and "output_manifest" from the result, and the question UUID; for a question with a push endpoint, the push subscription and the question UUID; for an asynchronous question, `None` and the question UUID
        """
        subscription, question_uuid = self._service.ask(
            service_id=self.id,
            input_values=input_values,
            input_manifest=input_manifest,
            children=children,
            subscribe_to_logs=subscribe_to_logs,
            allow_local_files=allow_local_files,
            save_diagnostics=save_diagnostics,
            question_uuid=question_uuid,
            parent_question_uuid=parent_question_uuid,
            originator_question_uuid=originator_question_uuid,
            originator=originator,
            push_endpoint=push_endpoint,
            asynchronous=asynchronous,
            retry_count=retry_count,
            timeout=timeout,
        )

        if push_endpoint or asynchronous:
            return subscription, question_uuid

        answer = self._service.wait_for_answer(
            subscription=subscription,
            handle_monitor_message=handle_monitor_message,
            record_events=record_events,
            timeout=timeout,
            maximum_heartbeat_interval=maximum_heartbeat_interval,
        )

        return answer, question_uuid

    def ask_multiple(
        self,
        *questions,
        raise_errors=True,
        max_retries=0,
        prevent_retries_when=None,
        max_workers=None,
        log_errors=True,
    ):
        """Ask the child multiple questions in parallel and wait for the answers. Each question should be provided as a
        dictionary of `Child.ask` keyword arguments. If `raise_errors` is `True`, an error is raised and no answers are
        returned if any of the individual questions raise an error; if it's `False`, answers are returned for all
        successful questions while errors are returned unraised for any failed ones.

        :param questions: any number of questions provided as dictionaries of arguments to the `Child.ask` method
        :param bool raise_errors: if `True`, an error is raised and no answers are returned if any of the individual questions raise an error; if `False`, answers are returned for all successful questions while errors are returned unraised for any failed ones
        :param int max_retries: retry any questions that failed up to this number of times (note: this will have no effect unless `raise_errors=False`)
        :param list(type)|None prevent_retries_when: prevent retrying any questions that fail with an exception type in this list (note: this will have no effect unless `raise_errors=False`)
        :param int|None max_workers: the maximum number of questions that can be asked at once; defaults to the lowest of {32, no. of CPUs + 4, and no. of questions} (see `concurrent.futures.ThreadPoolExecutor`)
        :param bool log_errors: if `True` and `raise_errors=False`, log any errors remaining once retries are exhausted
        :raise ValueError: if the maximum number of parallel questions is set too high
        :raise Exception: if any question raises an error if `raise_errors` is `True`
        :return list(dict|Exception, str): the answers or caught errors of the questions, and the question UUIDs (in the same order as asked)
        """
        prevent_retries_when = prevent_retries_when or []

        # Answers will come out of order, so use a dictionary to store them against their questions' original index.
        answers = {}
        n_questions = len(questions)
        max_workers = max_workers or min(32, (os.cpu_count() or 1) + 4, n_questions)
        logger.info("Asking %d questions with maximum %d threads.", n_questions, max_workers)

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            future_to_question_index_mapping = {}

            for i, question in enumerate(questions):
                # Add a question UUID if not set so retries have the same one.
                if "question_uuid" not in question:
                    question["question_uuid"] = str(uuid.uuid4())

                future = executor.submit(self.ask, **question)
                future_to_question_index_mapping[future] = i

            for i, future in enumerate(concurrent.futures.as_completed(future_to_question_index_mapping)):
                logger.info("%d of %d answers received.", i + 1, n_questions)
                question_index = future_to_question_index_mapping[future]

                try:
                    answers[question_index] = future.result()
                except Exception as e:
                    if raise_errors:
                        raise e

                    question_uuid = questions[question_index]["question_uuid"]
                    answers[question_index] = (e, question_uuid)

                    logger.error(
                        "Question %s failed. Run 'octue get-diagnostics gs://<diagnostics-cloud-path>/%s "
                        "--download-datasets' to get the crash diagnostics.",
                        question_uuid,
                        question_uuid,
                    )

        for retry in range(max_retries):
            failed_questions = self._get_failed_questions(
                questions,
                answers,
                prevent_retries_when,
                increment_retry_count=True,
            )

            if not failed_questions:
                break

            logger.info("%d questions failed - retrying %d of %d times.", len(failed_questions), retry + 1, max_retries)
            retried_answers = self.ask_multiple(*failed_questions.values(), raise_errors=False, log_errors=False)

            for question_index, answer in zip(failed_questions.keys(), retried_answers):
                answers[question_index] = answer

        if log_errors:
            # Check for failed questions after retries completed.
            failed_questions = self._get_failed_questions(questions, answers, prevent_retries_when)

            for question_index, question in failed_questions.items():
                logger.error(
                    "Question %s failed after %d retries (see below for error).",
                    question["question_uuid"],
                    max_retries,
                    exc_info=answers[question_index][0],
                )

        # Convert dictionary to list in asking order.
        return [answer[1] for answer in sorted(answers.items(), key=lambda item: item[0])]

    def _get_failed_questions(self, questions, answers, prevent_retries_when, increment_retry_count=False):
        """Get the questions that failed.

        :param list(dict) questions: the list of questions that were asked
        :param dict answers: a mapping of question index (i.e. position in the original list of questions) to question answer
        :param list(type)|None prevent_retries_when: prevent retrying any questions that fail with an exception type in this list (note: this will have no effect unless `raise_errors=False`)
        :param bool increment_retry_count: if `True`, increment the question retry count by 1
        :return dict: a mapping of failed question index (i.e. position in the original list of questions) to failed question
        """
        failed_questions = {}

        for question_index, answer in answers.items():
            if isinstance(answer[0], Exception) and type(answer[0]) not in prevent_retries_when:
                question = questions[question_index]

                if increment_retry_count:
                    question["retry_count"] = question.get("retry_count", 0) + 1

                failed_questions[question_index] = question

        return failed_questions



import concurrent.futures
import copy
import json
import logging
import os
import threading

from octue.cloud import storage
from octue.cloud.storage import GoogleCloudStorageClient
from octue.exceptions import InvalidInputException
from octue.mixins import Hashable, Identifiable, Metadata, Serialisable
from octue.resources.dataset import Dataset
from octue.utils.files import RegisteredTemporaryDirectory


logger = logging.getLogger(__name__)

# A lock to ensure only one thread can use the `Manifest.update_dataset_paths` method at a time. This avoids a race
# condition where a manifest shared across multiple questions in e.g. `Child.ask_multiple` has its dataset paths
# updated by more than one thread simultaneously while attempting to convert them to signed URLs. This lock should
# perhaps be applied to all methods where mutations are made to the manifest instead of just this single method.
update_dataset_paths_lock = threading.Lock()


class Manifest(Serialisable, Identifiable, Hashable, Metadata):
    """A representation of a manifest, which can contain multiple datasets This is used to manage all files coming into
    (or leaving), a data service for an analysis at the configuration, input or output stage.

    :param dict(str, octue.resources.dataset.Dataset|dict|str)|None datasets: a mapping of dataset names to `Dataset` instances, serialised datasets, or paths to datasets
    :param bool ignore_stored_metadata: if `True`, ignore any metadata stored for the manifest's datasets and datafiles locally or in the cloud
    :param str|None id: the UUID of the manifest (a UUID is generated if one isn't given)
    :param str|None name: an optional name to give to the manifest
    :return None:
    """

    _ATTRIBUTES_TO_HASH = ("datasets",)
    _METADATA_ATTRIBUTES = ("id",)

    # Paths to datasets are added to the serialisation in `Manifest.to_primitive`.
    _SERIALISE_FIELDS = (*_METADATA_ATTRIBUTES, "name")

    def __init__(self, datasets=None, ignore_stored_metadata=False, id=None, name=None):
        super().__init__(id=id, name=name)
        self._ignore_stored_metadata = ignore_stored_metadata
        self.datasets = self._instantiate_datasets(datasets or {})

    @classmethod
    def from_cloud(cls, cloud_path, ignore_stored_metadata=False):
        """Instantiate a manifest from a JSON serialisation of one in Google Cloud Storage.

        :param str cloud_path: full path to manifest in cloud storage (e.g. `gs://bucket_name/path/to/manifest.json`)
        :param bool ignore_stored_metadata: if `True`, ignore any metadata stored for the manifest's datasets and datafiles in the cloud
        :return octue.resources.manifest.Manifest:
        """
        serialised_manifest = json.loads(GoogleCloudStorageClient().download_as_string(cloud_path))

        return Manifest(
            id=serialised_manifest["id"],
            datasets=serialised_manifest["datasets"],
            ignore_stored_metadata=ignore_stored_metadata,
        )

    @property
    def all_datasets_are_in_cloud(self):
        """Do all the files of all the datasets of the manifest exist in the cloud?

        :return bool:
        """
        return all(dataset.all_files_are_in_cloud for dataset in self.datasets.values())

    def download(self, paths=None, download_all=True):
        """Download all datasets in the manifest. If no paths are provided, all datasets are downloaded to a temporary
        directory.

        :param dict|None paths: a mapping of dataset name to download directory path; if not provided, datasets are downloaded to temporary directories
        :param bool download_all: if `False` and `paths` is provided, only download the datasets specified in `paths`
        :return dict(str, str): the downloaded datasets mapped to the absolute paths of the directories they were downloaded into
        """
        if paths is None:
            download_all = True
            paths = {}

        datasets_directory = RegisteredTemporaryDirectory().name

        for name, dataset in self.datasets.items():
            download_path = paths.get(name)

            if not download_all and not download_path:
                logger.info("%r dataset download skipped as its download path wasn't specified.", name)
                continue

            paths[name] = dataset.download(local_directory=download_path or os.path.join(datasets_directory, name))

        return paths

    def update_dataset_paths(self, path_generator):
        """Update the path of each dataset according to the given path generator function. This method is thread-safe.

        :param callable path_generator: a function taking a `Dataset` as its only argument and returning the new path of the dataset
        :return None:
        """
        with update_dataset_paths_lock:
            for name, dataset in self.datasets.items():
                self.datasets[name].path = path_generator(dataset)

    def use_signed_urls_for_datasets(self):
        """Generate signed URLs for any cloud datasets in the manifest and use these as their paths instead of regular
        cloud paths. URLs will not be generated for any local datasets or datasets whose paths are already URLs
        (including those whose paths are already signed), making this method idempotent.

        :return None:
        """

        def signed_url_path_generator(dataset):
            if dataset.exists_in_cloud and storage.path.is_cloud_uri(dataset.path):
                return dataset.generate_signed_url()
            return dataset.path

        self.update_dataset_paths(signed_url_path_generator)
        logger.debug("Cloud paths (cloud URIs) for datasets replaced with signed URLs in %r.", self)

    def to_cloud(self, cloud_path):
        """Upload a manifest to a cloud location, optionally uploading its datasets into the same directory.

        :param str cloud_path: full path to cloud storage location to store manifest at (e.g. `gs://bucket_name/path/to/manifest.json`)
        :return None:
        """
        GoogleCloudStorageClient().upload_from_string(string=self.serialise(), cloud_path=cloud_path)

    def get_dataset(self, key):
        """Get a dataset by its key (as defined in the twine).

        :param str key:
        :return octue.resources.dataset.Dataset:
        """
        dataset = self.datasets.get(key, None)

        if dataset is None:
            raise InvalidInputException(
                f"Attempted to fetch unknown dataset {key!r} from Manifest. Allowable keys are: "
                f"{list(self.datasets.keys())}"
            )

        return dataset

    def prepare(self, data):
        """Prepare new manifest from a manifest_spec.

        :param dict data:
        :return Manifest:
        """
        if len(self.datasets) > 0:
            raise InvalidInputException("You cannot `prepare()` a manifest already instantiated with datasets")

        for key, dataset_specification in data["datasets"].items():
            # TODO generate a unique name based on the filter key, label datasets so that the label filters in the spec
            #  apply automatically and generate a description of the dataset
            self.datasets[key] = Dataset(path=key, ignore_stored_metadata=self._ignore_stored_metadata)

        return self

    def to_primitive(self):
        """Convert the manifest to a dictionary of primitives, converting its datasets into their paths for a
        lightweight serialisation.

        :return dict:
        """
        self_as_primitive = super().to_primitive()
        self_as_primitive["datasets"] = {}

        for name, dataset in self.datasets.items():
            if dataset._instantiated_from_files_argument:
                self_as_primitive["datasets"][name] = dataset.to_primitive()
            else:
                self_as_primitive["datasets"][name] = dataset.path

        return self_as_primitive

    def _instantiate_datasets(self, datasets):
        """Add the given datasets to the manifest, instantiating them if needed and giving them the correct path.
        There are several possible forms each dataset can come in:
        * Instantiated Dataset instance
        * A path to a dataset
        * Serialised form (a dictionary including a path key)

        The datasets can:
        * Including datafiles that already exist
        * Including datafiles that don't yet exist or are not possessed currently (e.g. future output locations or
          cloud files)

        :param dict(str, octue.resources.dataset.Dataset|dict|str) datasets: the datasets to add to the manifest
        :return dict:
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:
            return dict(executor.map(self._instantiate_dataset, copy.deepcopy(datasets).items()))

    def _instantiate_dataset(self, key_and_dataset):
        """Instantiate a dataset from multiple input formats.

        :param tuple(str, any) key_and_dataset:
        :return tuple(str, octue.resources.dataset.Dataset):
        """
        key, dataset = key_and_dataset

        if isinstance(dataset, Dataset):
            return (key, dataset)

        # If `dataset` is just a path to a dataset:
        if isinstance(dataset, str):
            return (key, Dataset(path=dataset, recursive=True, ignore_stored_metadata=self._ignore_stored_metadata))

        # If `dataset` is a cloud dataset and is represented as a dictionary including a "path" key:
        if storage.path.is_cloud_path(dataset.get("path", "")):
            return (
                key,
                Dataset(
                    path=dataset["path"],
                    files=dataset.get("files"),
                    recursive=True,
                    ignore_stored_metadata=self._ignore_stored_metadata,
                ),
            )

        return (key, Dataset(**dataset, ignore_stored_metadata=self._ignore_stored_metadata))

    def _set_metadata(self, metadata):
        """Set the manifest's metadata.

        :param dict metadata:
        :return None:
        """
        for attribute in self._METADATA_ATTRIBUTES:
            if attribute not in metadata:
                continue

            if attribute == "id":
                self._set_id(metadata["id"])
                continue

            setattr(self, attribute, metadata[attribute])



from google.cloud.bigquery import Client, QueryJobConfig, ScalarQueryParameter

from octue.cloud.events.validation import VALID_EVENT_KINDS


DEFAULT_FIELDS = (
    "`originator_question_uuid`",
    "`parent_question_uuid`",
    "`question_uuid`",
    "`kind`",
    "`event`",
    "`datetime`",
    "`uuid`",
    "`originator`",
    "`parent`",
    "`sender`",
    "`sender_type`",
    "`sender_sdk_version`",
    "`recipient`",
    "`other_attributes`",
)

BACKEND_METADATA_FIELDS = ("`backend`", "`backend_metadata`")


def get_events(
    table_id,
    question_uuid=None,
    parent_question_uuid=None,
    originator_question_uuid=None,
    kind=None,
    include_backend_metadata=False,
    tail=True,
    limit=1000,
):
    """Get Octue service events for a question from a Google BigQuery event store. Exactly one of the question UUID,
    parent question UUID, or originator question UUID must be provided:

    - When a question UUID is specified, only events from that question are retrieved
    - When a parent question UUID is specified, events from questions triggered by the same parent question are retrieved, not including the parent question's events
    - When an originator question UUID is specified, events for the entire tree of questions triggered by the originator question are retrieved, including the originator question's events

    When the limit is smaller than the total number of events, the default behaviour is to return the "tail" of the
    event stream for the question (the most recent n events for the question).

    :param str table_id: the full ID of the Google BigQuery table used as the event store e.g. "your-project.your-dataset.your-table"
    :param str|None question_uuid: the UUID of a question to get events for
    :param str|None parent_question_uuid: the UUID of a parent question to get the sub-question events for
    :param str|None originator_question_uuid: the UUID of an originator question get the full tree of events for
    :param str|None kind: the kind of event to get; if `None`, all event kinds are returned
    :param bool include_backend_metadata: if `True`, include the service backend metadata
    :param bool tail: if `True`, return the most recent events (where a limit applies); e.g. return the most recent 100 log records
    :param int limit: the maximum number of events to return
    :return list(dict): the events for the question; this will be empty if there are no events for the question
    """
    _validate_inputs(question_uuid, parent_question_uuid, originator_question_uuid, kind)

    if question_uuid:
        question_uuid_condition = "WHERE question_uuid=@relevant_question_uuid"
    elif parent_question_uuid:
        question_uuid_condition = "WHERE parent_question_uuid=@relevant_question_uuid"
    elif originator_question_uuid:
        question_uuid_condition = "WHERE originator_question_uuid=@relevant_question_uuid"

    if kind:
        event_kind_condition = [f"AND kind={kind!r}"]
    else:
        event_kind_condition = []

    # Make a shallow copy of the fields to query.
    fields = list(DEFAULT_FIELDS)

    if include_backend_metadata:
        fields.extend(BACKEND_METADATA_FIELDS)

    base_query = "\n".join(
        [
            f"SELECT {', '.join(fields)} FROM `{table_id}`",
            question_uuid_condition,
            *event_kind_condition,
        ]
    )

    if tail:
        # Order the inner query to get the most recent events at the top, then reorder the outer query to get the events
        # in natural order.
        query = "\n".join(
            [
                "SELECT * FROM (",
                base_query,
                "ORDER BY `datetime` DESC",
                "LIMIT @limit",
                ") ORDER BY `datetime` ASC",
            ]
        )
    else:
        query = "\n".join([base_query, "ORDER BY `datetime` ASC", "LIMIT @limit"])

    relevant_question_uuid = question_uuid or parent_question_uuid or originator_question_uuid

    job_config = QueryJobConfig(
        query_parameters=[
            ScalarQueryParameter("relevant_question_uuid", "STRING", relevant_question_uuid),
            ScalarQueryParameter("limit", "INTEGER", limit),
        ]
    )

    client = Client()
    query_job = client.query(query, job_config=job_config)
    result = query_job.result()

    if result.total_rows == 0:
        return []

    events = [_deserialise_event(event) for event in result]
    return _unflatten_events(events)


def _validate_inputs(question_uuid, parent_question_uuid, originator_question_uuid, kind):
    """Check that only one of `question_uuid`, `parent_question_uuid`, or `originator_question_uuid` are provided and
    that the `kind` parameter is a valid event kind.

    :param str|None question_uuid: the UUID of a question to get events for
    :param str|None parent_question_uuid: the UUID of a parent question to get the sub-question events for
    :param str|None originator_question_uuid: the UUID of an originator question get the full tree of events for
    :param str|None kind: the kind of event to get; if `None`, all event kinds are returned
    :raise ValueError: if more than one of `question_uuid`, `parent_question_uuid`, or `originator_question_uuid` are provided or the `kind` parameter is invalid
    :return None:
    """
    question_uuid_inputs = (bool(question_uuid), bool(parent_question_uuid), bool(originator_question_uuid))

    if sum(question_uuid_inputs) != 1:
        raise ValueError(
            "One and only one of `question_uuid`, `parent_question_uuid`, or `originator_question_uuid` must be "
            "provided."
        )

    if kind and kind not in VALID_EVENT_KINDS:
        raise ValueError(f"`kind` must be one of {VALID_EVENT_KINDS!r}; received {kind!r}.")


def _deserialise_event(event):
    """Deserialise an event from the event store:

    - Convert "null" to `None` in the `parent_question_uuid` field
    - Convert string-cast booleans and integers to `bool` and `int` types

    :param google.cloud.bigquery.table.Row event: a serialised event from the event store
    :return dict: the deserialised event
    """
    event = dict(event)

    if event["parent_question_uuid"] == "null":
        event["parent_question_uuid"] = None

    # Convert string-cast attributes back to `bool` or `int`.
    other_attributes = event["other_attributes"]
    other_attributes["retry_count"] = int(other_attributes.get("retry_count"))

    if other_attributes.get("forward_logs"):
        other_attributes["forward_logs"] = bool(int(other_attributes.get("forward_logs")))

    return event


def _unflatten_events(events):
    """Convert the events and attributes from the flat structure of the BigQuery table into the nested structure of the
    service communication schema.

    :param list(dict) events: flattened events
    :return list(dict): unflattened events
    """
    for event in events:
        event["event"]["kind"] = event.pop("kind")

        event["attributes"] = {
            "originator_question_uuid": event.pop("originator_question_uuid"),
            "parent_question_uuid": event.pop("parent_question_uuid"),
            "question_uuid": event.pop("question_uuid"),
            "datetime": event.pop("datetime").isoformat(),
            "uuid": event.pop("uuid"),
            "originator": event.pop("originator"),
            "parent": event.pop("parent"),
            "sender": event.pop("sender"),
            "sender_type": event.pop("sender_type"),
            "sender_sdk_version": event.pop("sender_sdk_version"),
            "recipient": event.pop("recipient"),
            **event.pop("other_attributes"),
        }

    return events



import importlib.metadata
from abc import abstractmethod

from octue.mixins.hashable import Hashable


class Metadata:
    _METADATA_ATTRIBUTES = tuple()

    @property
    @abstractmethod
    def metadata_path(self):
        """Get the path to the instance's local metadata file if it has one.

        :return str|None:
        """
        pass

    @property
    def metadata_hash_value(self):
        """Get the hash of the instance's metadata, not including its ID.

        :return str:
        """
        return self._metadata_hash_value()

    def metadata(self, include_id=True, include_sdk_version=True, **kwargs):
        """Get the instance's metadata in primitive form. The metadata is the set of attributes included in the class
        variable `self._METADATA_ATTRIBUTES`.

        :param bool include_id: if `True`, include the ID of the instance if it is included in `self._METADATA_ATTRIBUTES`
        :param bool include_sdk_version: if `True`, include the `octue` version that instantiated the instance
        :param kwargs: any kwargs to use in an overridden `self.metadata` method
        :return dict:
        """
        metadata = {name: getattr(self, name) for name in self._METADATA_ATTRIBUTES}

        if include_sdk_version:
            metadata["sdk_version"] = importlib.metadata.version("octue")

        if not include_id and "id" in metadata:
            del metadata["id"]

        return metadata

    def _metadata_hash_value(self, **kwargs):
        """Get the hash of the instance's metadata, not including its ID. Override this method to change what kwargs
        `self.metadata` gets.

        :param kwargs: any kwargs to use in an overridden `self.metadata` method when calculating the metadata hash value
        :return str:
        """
        return Hashable.hash_non_class_object(self.metadata(include_id=False, include_sdk_version=False, **kwargs))

    @abstractmethod
    def _set_metadata(self, metadata):
        """Set the instance's metadata.

        :param dict metadata:
        :return None:
        """
